# -*- coding:utf-8; eval: (blacken-mode) -*-
import logging
import numpy as np
from scipy.linalg import signm

from ._core import Action, MASK_NONE, MASK_TERMINAL, MASK_VACUUM

__all__ = ["Core", "MASK_NONE", "MASK_TERMINAL", "MASK_VACUUM"]

_log = logging.getLogger(__name__)


def _action_property(name, freeze=True):
    def fget(self):
        return getattr(self, "_" + name)

    def fset(self, value):
        if self._frozen and freeze:
            raise RuntimeError("Call reset() before modifications")
        v = getattr(self, "_" + name)
        if isinstance(v, np.ndarray):
            raise AttributeError("Arrays are not assignable")
        value = type(v)(value)
        if self._action is not None and hasattr(self._action, name):
            setattr(self._action, name, value)
        setattr(self, "_" + name, value)

    return property(fget, fset)


class Core:
    def __init__(self, Lx=1, Ly=1, eta=0, D=1, nx=1, ny=1):
        self._Lx = float(Lx)
        self._Ly = float(Ly)
        self._eta = complex(eta)
        self._D = complex(D)
        self.set_shape(nx, ny)

    def set_shape(self, nx=1, ny=1):
        self._frozen = False
        self._shape = (nx, ny)
        self._omega = complex(0)
        self._mask = np.zeros((nx, ny), dtype=np.uint8)
        self._U = np.zeros((nx, ny, 4, 4, 4), dtype=complex)
        self._Omega = np.zeros((nx, ny, 4, 4), dtype=complex)
        self._action = None
        self._U[...] = np.eye(4)

    def init_U(self):
        self._U[1:, :, 2, :, :] = np.linalg.inv(self._U[:-1, :, 0, :, :])
        self._U[:, 1:, 3, :, :] = np.linalg.inv(self._U[:, :-1, 1, :, :])

    def fix_terminals(self, Phi):
        m = self._mask == MASK_TERMINAL
        tau3 = np.diag([1, 1, -1, -1])
        for i, j in zip(*np.where(m)):
            M = self._Omega[i, j] + 1j * self._omega * tau3
            Q = signm(-1j * M)
            g = np.linalg.solve(np.eye(2) + Q[:2, :2], Q[:2, 2:])
            gt = np.linalg.solve(np.eye(2) - Q[2:, 2:], Q[2:, :2])
            Phi[i, j, 0] = g
            Phi[i, j, 1] = gt

    def reset(self):
        self._action = None
        self._frozen = False
        self._U.flags.writeable = True
        self._mask.flags.writeable = True

    def update(self):
        self.init_U()
        self._action = Action(
            self.Lx, self.Ly, self.eta, self.D, self._mask, self._U, self._Omega
        )
        self._action.omega = self._omega
        self._frozen = True
        self._U.flags.writeable = False
        self._mask.flags.writeable = False

    def _ensure_action(self):
        if self._action is None:
            self.update()

    @property
    def x(self):
        return np.linspace(-self.Lx / 2, self.Lx / 2, self._shape[0])

    @property
    def y(self):
        return np.linspace(-self.Ly / 2, self.Ly / 2, self._shape[1])

    def eval(self, Q):
        self._ensure_action()
        return self._action.eval(Q)

    def grad(self, Q):
        self._ensure_action()
        return self._action.grad(Q)

    def hess(self, Q):
        self._ensure_action()
        return self._action.hess(Q)

    def hess_mul(self, Q, dQ):
        self._ensure_action()
        return self._action.hess_mul(Q, dQ)

    def grad_A(self, Q):
        self._ensure_action()
        return self._action.grad_A(Q)

    def grad_Omega(self, Q):
        self._ensure_action()
        return self._action.grad_Omega(Q)

    U = _action_property("U")
    Omega = _action_property("Omega")
    mask = _action_property("mask")
    omega = _action_property("omega", freeze=False)
    hess_computed = _action_property("hess_computed")
    shape = property(fget=lambda self: self._shape)
    Lx = _action_property("Lx")
    Ly = _action_property("Ly")
    D = _action_property("D")
    eta = _action_property("eta")

    @property
    def hess_computed(self):
        return self._action.hess_computed if self._action is not None else False

    @property
    def Delta(self):
        return _DeltaArray(self.Omega)

    @property
    def Ux(self):
        return self.U[:, :, 0, :, :]

    @property
    def Uy(self):
        return self.U[:, :, 1, :, :]


class _DeltaArray:
    def __init__(self, Omega):
        self._Omega = Omega

    def __getitem__(self, ix):
        z = self._Omega[:, :, :2, 2:][ix] / 1j
        if isinstance(z, np.ndarray):
            z.flags.writeable = False
        return z

    def __setitem__(self, ix, value):
        self._Omega[:, :, :2, 2:][ix] = 1j * value
        self._Omega[:, :, 2:, :2].transpose(0, 1, 3, 2)[ix] = 1j * np.conjugate(value)

    @property
    def A(self):
        return self[...]

    def copy(self):
        return self[...].copy()

    @property
    def shape(self):
        return self._Omega.shape[:-2] + (2, 2)
